--#########################################################
-- PALGen
--#########################################################

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

--#########################################################
-- entity
--#########################################################

entity PALGen is
	generic
	(
		CLK_MHZ: integer := 100 -- multiple of 12.5 MHz
	);
	port
	(
		clk  : in  std_logic;
		pixel: out std_logic;
		porch: out std_logic;
		color: out std_logic;
		hpos : out std_logic_vector(9 downto 0);
		vpos : out std_logic_vector(9 downto 0)
	);
end;

--#########################################################
-- architecture
--#########################################################

architecture arch of PALGen is

	-----------------------------------------------
	-- PAL attributes
	-----------------------------------------------

--	constant PAL_KHZ  : integer := 12500;
--	constant HPOS_MAX : integer := 40;
--	constant VPOS_MAX : integer := 14;
--	constant HSYN_LEN : integer :=  2;
--	constant HPIX_BEG : integer := 10;
--	constant HPIX_END : integer := 38;

	constant PAL_KHZ  : integer := 12500;
	constant HPOS_MAX : integer := 800;
	constant VPOS_MAX : integer := 313; -- = 625/2
	constant HSYN_LEN : integer :=  50;
	constant HPIX_BEG : integer := 120;
	constant HPIX_END : integer := 760;

	-----------------------------------------------
	-- clock divider
	-----------------------------------------------

--	constant TIMER_MAX: integer := 2;
--	constant TIMER_MAX: integer := 8;

	constant TIMER_MAX: integer := CLK_MHZ*1000/PAL_KHZ;

	signal timer_reg: integer range 0 to TIMER_MAX-1 := 0;
	signal timer_nxt: integer range 0 to TIMER_MAX-1;

	signal cen_reg: std_logic := '0';
	signal cen_nxt: std_logic;

	signal cen: std_logic;

	-----------------------------------------------
	-- horz,ver counters:
	-----------------------------------------------

	signal hpos_reg: integer range 0 to HPOS_MAX-1 := 0;
	signal hpos_nxt: integer range 0 to HPOS_MAX-1;

	signal vpos_reg: integer range 0 to VPOS_MAX-1 := 0;
	signal vpos_nxt: integer range 0 to VPOS_MAX-1;

	signal hbeg_reg: std_logic := '0';
	signal hbeg_nxt: std_logic;

	signal vbeg_reg: std_logic := '0';
	signal vbeg_nxt: std_logic;

	signal even_reg: std_logic := '0';
	signal even_nxt: std_logic;

	signal hpos_val: integer range 0 to HPOS_MAX-1;
	signal vpos_val: integer range 0 to VPOS_MAX-1;
	signal hbeg: std_logic;
	signal vbeg: std_logic;
	signal even: std_logic;

	-----------------------------------------------
	-- position marker
	--
	-- ##----##----##    BEG1/2,MID1/2,END1/2
	-- --####--######    BLK1/2,BLK3/4
	-- ---#########--    PIX1/2
	-----------------------------------------------

	constant HPOS_BEG1: integer :=              0;
	constant HPOS_BEG2: integer := HSYN_LEN   + 0;

	constant HPOS_MID1: integer := HPOS_MAX/2 - HSYN_LEN/2;
	constant HPOS_MID2: integer := HPOS_MAX/2 + HSYN_LEN/2;

	constant HPOS_END1: integer := HPOS_MAX   - HSYN_LEN;
	constant HPOS_END2: integer := HPOS_MAX   + 0;

	constant HPOS_BLK1: integer :=              HSYN_LEN;
	constant HPOS_BLK2: integer := HPOS_MAX/2 - HSYN_LEN/2;
	constant HPOS_BLK3: integer := HPOS_MAX/2 + HSYN_LEN/2;
	constant HPOS_BLK4: integer := HPOS_MAX   + 0;

	constant HPOS_PIX1: integer := HPIX_BEG;
	constant HPOS_PIX2: integer := HPIX_END;


	signal hpos_begin : std_logic;
	signal hpos_middle: std_logic;
	signal hpos_end   : std_logic;
	signal hpos_block1: std_logic;
	signal hpos_block2: std_logic;
	signal hpos_pixels: std_logic;

	-----------------------------------------------
	-- FSM
	-----------------------------------------------

	type FSM is
	(
		S_ODD_VERT_SYNC,
		S_ODD_POST_PULSES1,
		S_ODD_POST_PULSES2,
		S_ODD_FIELD,
		S_ODD_PRE_PULSES1,
		S_ODD_PRE_PULSES2,

		S_EVEN_VERT_SYNC,
		S_EVEN_POST_PULSES,
		S_EVEN_FIELD,
		S_EVEN_PRE_PULSES
	);
	
	signal state_reg: FSM := S_ODD_VERT_SYNC;
	signal state_nxt: FSM;

	signal porch_reg: std_logic := '0';
	signal porch_nxt: std_logic;

	signal color_reg: std_logic := '0';
	signal color_nxt: std_logic;

begin

	-----------------------------------------------
	-- clock divider
	-----------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			timer_reg <= timer_nxt;
			cen_reg   <= cen_nxt;
		end if;
	end process;

	process
	(
		timer_reg,
		cen_reg,
		hbeg_reg,
		hbeg_reg,
		even_reg
	)
	begin
		timer_nxt <= timer_reg;
		cen_nxt   <= cen_reg;

		if timer_reg /= TIMER_MAX-1 then
			timer_nxt <= timer_reg + 1;
			cen_nxt   <= '0';
		else
			timer_nxt <= 0;
			cen_nxt   <= '1';
		end if;

	end process;

	cen <= cen_reg;

	-----------------------------------------------
	-- horz,ver counters
	-----------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			if cen = '1' then
				hpos_reg <= hpos_nxt;
				vpos_reg <= vpos_nxt;
				hbeg_reg <= hbeg_nxt;
				vbeg_reg <= vbeg_nxt;
				even_reg <= even_nxt;
			end if;
		end if;
	end process;

	process
	(
		hpos_reg,
		hpos_reg,
		hbeg_reg,
		hbeg_reg,
		even_reg
	)
	begin
		hpos_nxt <= hpos_reg;
		vpos_nxt <= vpos_reg;
		hbeg_nxt <= hbeg_reg;
		vbeg_nxt <= vbeg_reg;
		even_nxt <= even_reg;


		hbeg_nxt <= '0';
		vbeg_nxt <= '0';

		if hpos_reg < HPOS_MAX-1 then
			hpos_nxt <= hpos_reg + 1;
		else
			hpos_nxt <= 0;
			hbeg_nxt <= '1';

			if (even_reg = '0' and vpos_reg < VPOS_MAX-1) 
			or (even_reg = '1' and vpos_reg < VPOS_MAX-2) then
				vpos_nxt <= vpos_reg + 1;
			else
				vpos_nxt <= 0;
				vbeg_nxt <= '1';
			end if;

		end if;

	end process;

	hpos_val <= hpos_reg;
	vpos_val <= vpos_reg;
	hbeg     <= hbeg_reg;
	vbeg     <= vbeg_reg;
	even     <= even_reg;

	-----------------------------------------------
	-- position marker
	-----------------------------------------------

	process(hpos_val)
	begin

		-- begin,middle,end markers
		if hpos_val >= HPOS_BEG1 and hpos_val < HPOS_BEG2 then
			hpos_begin <= '1';
		else
			hpos_begin <= '0';
		end if;
		
		if hpos_val >= HPOS_MID1 and hpos_val < HPOS_MID2 then
			hpos_middle <= '1';
		else
			hpos_middle <= '0';
		end if;
		
		if hpos_val >= HPOS_END1 and hpos_val < HPOS_END2 then
			hpos_end <= '1';
		else
			hpos_end <= '0';
		end if;


		-- block1/2 markers
		if hpos_val >= HPOS_BLK1 and hpos_val < HPOS_BLK2 then
			hpos_block1 <= '1';
		else
			hpos_block1 <= '0';
		end if;

		if hpos_val >= HPOS_BLK3 and hpos_val < HPOS_BLK4 then
			hpos_block2 <= '1';
		else
			hpos_block2 <= '0';
		end if;


		-- field pixels markers
		if hpos_val >= HPOS_PIX1 and hpos_val < HPOS_PIX2 then
			hpos_pixels <= '1';
		else
			hpos_pixels <= '0';
		end if;

	end process;

	-----------------------------------------------
	-- FSM
	-----------------------------------------------

	process(clk)
	begin
		if rising_edge(clk) then
			if cen = '1' then
				state_reg <= state_nxt;
				porch_reg <= porch_nxt;
				color_reg <= color_nxt;
			end if;
		end if;
	end process;

	process
	(
		state_reg,
		porch_reg,
		color_reg,
		hpos_val,
		hpos_val,
		hbeg,
		vbeg,
		hpos_begin,
		hpos_middle,
		hpos_end,
		hpos_block1,
		hpos_block2,
		hpos_pixels
	)
	begin
		state_nxt <= state_reg;
		porch_nxt <= porch_reg;
		color_nxt <= color_reg;


		case state_reg is

			when S_ODD_VERT_SYNC => 

					-- /////////////////////
					-- odd frame: line 1,2
					-- ----#----# ----#----#
					-- \\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 2 then
						state_nxt <= S_ODD_POST_PULSES1;
					end if;

					porch_nxt <= '0';
					color_nxt <= '0';

					if hpos_middle = '1' or hpos_end = '1' then
						porch_nxt <= '1';
					end if;

			when S_ODD_POST_PULSES1 =>

					-- /////////////////
					-- odd frame: line 3
					-- ----#-####
					-- \\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 3 then
						state_nxt <= S_ODD_POST_PULSES2;
					end if;

					porch_nxt <= '0';
					color_nxt <= '0';

					if hpos_middle = '1' or hpos_block2 = '1' then
						porch_nxt <= '1';
					end if;

			when S_ODD_POST_PULSES2 =>

					-- /////////////////////
					-- odd frame: line 4,5
					-- -####-#### -####-####
					-- \\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 5 then
						state_nxt <= S_ODD_FIELD;
					end if;

					porch_nxt <= '1';
					color_nxt <= '0';

					if hpos_begin = '1' or hpos_middle = '1' then
						porch_nxt <= '0';
					end if;

			when S_ODD_FIELD =>

					-- //////////////////////////
					-- odd frame: line 6,..n-3
					-- -######### ...  -#########
					-- \\\\\\\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = VPOS_MAX-3 then
						state_nxt <= S_ODD_PRE_PULSES1;
					end if;

					porch_nxt <= '1';
					color_nxt <= '0';

					if hpos_begin = '1' then
						porch_nxt <= '0';
					end if;

					if hpos_pixels = '1' then
						color_nxt <= '1';
					end if;

			when S_ODD_PRE_PULSES1 =>

					-- ///////////////////////
					-- odd frame: line n-2,n-1
					-- -####-#### -####-####
					-- \\\\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = VPOS_MAX-1 then
						state_nxt <= S_ODD_PRE_PULSES2;
					end if;

					porch_nxt <= '0';
					color_nxt <= '0';

					if hpos_block1 = '1' or hpos_block2 = '1' then
						porch_nxt <= '1';
					end if;

			when S_ODD_PRE_PULSES2 =>

					-- /////////////////////
					-- odd frame: line n-1,n
					-- -####----#
					-- \\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 0 then
						state_nxt <= S_EVEN_VERT_SYNC;
					end if;

					porch_nxt <= '0';
					color_nxt <= '0';

					if hpos_block1 = '1' or hpos_end = '1' then
						porch_nxt <= '1';
					end if;


			when S_EVEN_VERT_SYNC =>

					-- /////////////////////
					-- even frame: line 1,2
					-- ----#----# ----#----#
					-- \\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 2 then
						state_nxt <= S_EVEN_POST_PULSES;
					end if;

					porch_nxt <= '0';
					color_nxt <= '0';

					if hpos_block1 = '1' or hpos_middle = '1' then
						porch_nxt <= '1';
					end if;

			when S_EVEN_POST_PULSES =>

					-- /////////////////////
					-- odd frame: line 3,4
					-- -####-#### -####-####
					-- \\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 4 then
						state_nxt <= S_EVEN_FIELD;
					end if;

					porch_nxt <= '1';
					color_nxt <= '0';

					if hpos_begin = '1' or hpos_middle = '1' then
						porch_nxt <= '1';
					end if;

			when S_EVEN_FIELD =>

					-- /////////////////////////
					-- odd frame: line 6,..n-3
					-- -######### ... -#########
					-- \\\\\\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = VPOS_MAX-3 then
						state_nxt <= S_EVEN_PRE_PULSES;
					end if;

					porch_nxt <= '1';
					color_nxt <= '0';

					if hpos_begin = '1' then
						porch_nxt <= '0';
					end if;

					if hpos_pixels = '1' then
						color_nxt <= '1';
					end if;

			when S_EVEN_PRE_PULSES =>

					-- ////////////////////////////////
					-- odd frame: line n-2,n-1,n
					-- -####-#### -####-#### -####-####
					-- \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\

					if hbeg = '1' and vpos_val = 0 then
						state_nxt <= S_ODD_VERT_SYNC;
					end if;

					porch_nxt <= '1';
					color_nxt <= '0';

					if hpos_begin = '1' or hpos_middle = '1' then
						porch_nxt <= '1';
					end if;

		end case;

	end process;

	-----------------------------------------------
	-- I/O
	-----------------------------------------------

	pixel <= cen;
	porch <= porch_reg;
	color <= color_reg;

	hpos <= std_logic_vector(to_unsigned(hpos_val,10));
	vpos <= std_logic_vector(to_unsigned(vpos_val,10));

end;

--#########################################################
-- end
--#########################################################
